import numpy as np
import copy

def inicializarLaberinto():
    laberinto = np.zeros([10, 10])
    paredes = np.array([[0, 2], [0, 7],[1, 0], [1, 2], [1, 5], [1, 6], [1, 8],[2,6],[2,8],[3,1],[3,4],[3,5],[3,6],[4,2],[4,3],[4,7],[5,5],[5,7],[6,0],[6,3],[6,4],[6,7],[6,9],[7,1],[7,2],[7,8],[7,9],[8,2],[8,4],[8,5]])
    laberinto[paredes[:, 0], paredes[:, 1]] = np.inf
    return laberinto


def encolar(q, sol):
    i=0
    while i<len(q) and sol[0]>=q[i][0]:
        i+=1
    q = q[:i] + [sol] + q[i:]
    return q

def desencolar(q):
    return [q[0], q[1:]]

def crearColaVacia():
    return []

def esColaVacia(q):
    return q == []

def GenerarTuplaVacia(laberinto):
    N = len(laberinto)
    tupla = np.copy(laberinto)
    f=0
    c=0
    longitud = 1 + N-f + N-c
    tupla[f,c] = 1
    return [longitud,tupla,f,c]

def SolucionInicial(laberinto):
    sol = GenerarTuplaVacia(laberinto)
    sol[0] = np.sum(laberinto == 0)
    return sol

def esSolucion(sol):
    n = len(sol[1])
    return sol[2] == n-1 and sol[3] == n-1

def esMejor(sol1,sol2):
    return sol1[0] < sol2[0]

def esFactible(laberinto,f,c):
    if f>=0 and f<np.size(laberinto,0) and c>=0 and c<np.size(laberinto,1):
        return laberinto[f][c] == 0
    else:
        return False

def asignar(sol,f,c):
    N = len(laberinto)
    tupla = sol[1]
    anotacion = tupla[sol[2],sol[3]]
    tupla[f,c] = anotacion + 1
    longitud = anotacion + 1 + N-f + N-c
    return [longitud, tupla, f, c]

def compleccionesFactibles(sol):
    desplazamientos = np.array([[1, 0], [0, 1], [-1, 0], [0, -1]])
    hijos = []
    for i in range(len(desplazamientos)):
        f = sol[2] + desplazamientos[i][0]
        c = sol[3] + desplazamientos[i][1]
        if esFactible(sol[1],f,c):
            hijo = copy.deepcopy(sol)
            hijo = asignar(hijo,f,c)
            hijos.append(hijo)
    return hijos


def RamificacionYPoda(laberinto):
    solMejor = SolucionInicial(laberinto) # Sólo Cota. No es solución
    sol = GenerarTuplaVacia(laberinto) # Tupla parcial inicialmente vacía
    q = crearColaVacia()
    q = encolar(q, sol) # se añade la sol inicial
    while not esColaVacia(q):
        [sol, q] = desencolar(q) # Sol es el nodo de prioridad max
        if esSolucion(sol):
            if esMejor(sol, solMejor):
                solMejor = sol # Se actualiza SolMejor
        else:
            if esMejor(sol, solMejor): # Criterio de poda
                hijos = compleccionesFactibles(sol)
                for hijo in hijos:
                    if esMejor(hijo, solMejor):
                        q = encolar(q,hijo)
    return solMejor

# Prog Ppal:
laberinto = inicializarLaberinto()
sol = RamificacionYPoda(laberinto)
print(sol[1])
