import numpy as np
import copy

def encolar(q, sol):
    i=0
    while i<len(q) and sol[0]>=q[i][0]:
        i+=1
    q = q[:i] + [sol] + q[i:]
    return q

def desencolar(q):
    return [q[0], q[1:]]

def crearColaVacia():
    return []

def esColaVacia(q):
    return q == []

def  SolucionInicial(laberinto):
    n = len(laberinto)
    sol = [n*n, copy.deepcopy(laberinto), n-1,n-1]
    return sol

def calculaCoste(sol):
    laberinto = sol[1]
    n = len(laberinto)
    fila = sol[2]
    columna = sol[3]
    costeReal = laberinto[fila,columna]
    costeEstimado = (n-1 - fila) + (n-1 - columna)
    sol[0] = costeReal + costeEstimado
    return sol

def GenerarTuplaVacia(laberinto):
    # La solución es una lista con esta pinta: [Coste Solucion, Solucion en Laberinto, Fila, Columna]
    laberintoCp = copy.deepcopy(laberinto)
    laberintoCp[0,0] = 1
    sol = [0, laberintoCp, 0, 0]
    sol = calculaCoste(sol)
    return sol

def esSolucion(sol,n):
    return sol[2] == n - 1 and sol[3] == n - 1

def esMejor(sol1,sol2):
    return sol1[0] < sol2[0]

def esFactible(laberinto,f,c):
    n = len(laberinto)
    if f < 0 or f >= n or c < 0 or c >= n:
        return False
    else:
        return laberinto[f,c] == 0

def hijosFactibles(sol):
    desplazamientos = np.array([[1, 0], [-1 , 0], [0, 1],[0, -1] ], dtype = int)
    hijos = []
    filaActual = sol[2]
    colActual = sol[3]
    laberinto = sol[1]
    for i in range(len(desplazamientos)):
        filaCandidata = filaActual + desplazamientos[i,0]
        colCandidata = colActual + desplazamientos[i,1]
        if esFactible(laberinto,filaCandidata,colCandidata):
            laberintoCp = copy.deepcopy(laberinto)
            laberintoCp[filaCandidata,colCandidata] = laberinto[filaActual,colActual] + 1
            hijo = [0, laberintoCp, filaCandidata, colCandidata]
            hijo = calculaCoste(hijo)
            hijos.append(hijo)
    return hijos

def RamificacionYPoda(laberinto):
    n = len(laberinto)
    solMejor = SolucionInicial(laberinto) # Algoritmo voraz
    sol = GenerarTuplaVacia(laberinto) # Tupla parcial inicialmente vacía
    q = crearColaVacia()
    q = encolar(q, sol) # se añade la sol inicial
    while not esColaVacia(q):
        [sol, q] = desencolar(q) # Sol es el nodo de prioridad max
        if esSolucion(sol,n):
            if esMejor(sol, solMejor):
                solMejor = sol # Se actualiza SolMejor
        else:
            if esMejor(sol, solMejor): # Criterio de poda
                hijos = hijosFactibles(sol)
                for hijo in hijos:
                    if esMejor(hijo, solMejor):
                        q = encolar(q,hijo)
    return solMejor

def inicializarLaberinto():
    laberinto = np.zeros([10, 10])
    paredes = np.array([[0, 2], [0, 7],[1, 0], [1, 2], [1, 5], [1, 6], [1, 8],[2,6],[2,8],[3,1],[3,4],[3,5],[3,6],[4,2],[4,3],[4,7],[5,5],[5,7],[6,0],[6,3],[6,4],[6,7],[6,9],[7,1],[7,2],[7,8],[7,9],[8,2],[8,4],[8,5]])
    laberinto[paredes[:, 0], paredes[:, 1]] = np.inf
    return laberinto




# Prog Ppal:
laberinto = inicializarLaberinto()
sol = RamificacionYPoda(laberinto)
print(sol)