import numpy as np
import copy


def ordenar(coleccion,criterio):
    for j in range(1, len(coleccion)):
        clave = np.copy(coleccion[j][:])
        # Insertar la clave en la secuencia ordenada:
        i = j - 1
        while i >= 0 and coleccion[i][criterio] < clave[criterio]:
            coleccion[i + 1][:] = np.copy(coleccion[i][:])
            i = i - 1
        coleccion[i + 1][:] = np.copy(clave[:])
    return coleccion

def GenerarTuplaVacia(n):
    # La solución es una colección formada por: [prioridad, tupla, longitud, Valor Func Coste]
    solucion = [0, np.zeros(n,dtype=int), 0, 0]
    solucion = calculaCota(solucion, objetos, W)
    return solucion

def SolucionInicial(W,objetos):
    n = len(objetos)
    tupla = np.zeros(n,dtype=int)
    peso = 0
    i = 0
    while i < n and peso < W:
        if peso + objetos[i,1] <= W:
            peso += objetos[i,1]
            tupla[i] = 1
        i += 1
    # La solución es una colección formada por: [prioridad, tupla, longitud, Valor Func Coste]
    solucion = [0, tupla, n, 0] # Longitud == n ==> solución completa
    solucion = calculaCota(solucion, objetos, W) # Para soluciones completas, la cota coincide con el coste
    return solucion


def calculaCota(solucion,objetos,W):
    tupla = solucion[1]
    longitud = solucion[2]
    valorObjetos = np.sum(np.multiply(tupla[0:longitud],objetos[0:longitud, 0]))
    if longitud == len(tupla):
        solucion[3] = valorObjetos
    else:
        pesoObjetos = np.sum(np.multiply(tupla[0:longitud], objetos[0:longitud, 1]))
        solucion[3] = valorObjetos + (W - pesoObjetos)*objetos[longitud,2]
    solucion[0] = 1 / solucion[3] # Prioridad
    return solucion

def esSolucion(sol,n):
    return sol[2] == n

def esMejor(sol1, sol2):
    return sol1[3] > sol2[3]

def asignar(sol,valor,W,objetos):
    longitud = sol[2]
    sol[1][longitud] = valor # Se asigna nuevo valor
    sol[2] += 1 # Se aumenta la longitud
    sol = calculaCota(sol, objetos, W)
    return sol

def complecciones(sol, W, objetos):
    # Dos posibilidades (se incluye o no el objeto en la solucion):
    hijo1 = copy.deepcopy(sol) # Genera una copia de listas y sublistas
    hijo2 = copy.deepcopy(sol)
    hijo1 = asignar(hijo1,0,W,objetos)
    hijo2 = asignar(hijo2,1,W,objetos)
    return [hijo1, hijo2]


def esFactible(sol,W,objetos):
    long = sol[2]
    peso = np.sum(np.multiply(sol[1][0:long],objetos[0:long,1]))
    return peso <= W

def encolar(q, sol):
    i=0
    while i<len(q) and sol[0]>=q[i][0]:
        i+=1
    q = q[:i] + [sol] + q[i:]
    return q

def desencolar(q):
    return [q[0], q[1:]]

def crearColaVacia():
    return []

def esColaVacia(q):
    return q == []


def RamificacionYPoda(W,objetos):
    n = len(objetos)
    solMejor = SolucionInicial(W,objetos) # Algoritmo voraz
    sol = GenerarTuplaVacia(n) # Tupla parcial inicialmente vacía
    q = crearColaVacia()
    q = encolar(q, sol) # se añade la sol inicial
    while not esColaVacia(q):
        [sol, q] = desencolar(q) # Sol es el nodo de prioridad max
        if esSolucion(sol,n):
            if esMejor(sol, solMejor):
                solMejor = sol # Se actualiza SolMejor
        else:
            if esMejor(sol, solMejor): # Criterio de poda
                hijos = complecciones(sol,W,objetos)
                for hijo in hijos:
                    if esFactible(hijo,W,objetos) and esMejor(hijo, solMejor):
                        q = encolar(q,hijo)
    return solMejor



# Pog Ppal:
# datos:
W = 15
objetos = np.array([[12,6],[10,4],[10,2],[18,9]],dtype=float) # [v, w]
objetos = np.insert(objetos,2,0,axis=1) # añadimos una nueva columna para guardar v/w:
objetos[:,2] = objetos[:,0] / objetos[:,1] # [v, w, v/w]
objetos = ordenar(objetos,2) # Los ordenamos por v/w decreciente
# Resolución del problema:
sol = RamificacionYPoda(W,objetos)
print('La mejor solucion es:',sol[1])
print('Con valor:',sol[3])